/* (C) Copyright 1984,85,86,87 Walter L. Peacock   All Rights Reserved    */
/*  G E T B T R E E . C   Routines to read the index node records. */

/*----------------------------------------------------------------------------
 * getroot() returns the pointer value to the B+tree node.   The
 * root node is the entry point into the B+tree.  The location of the root node
 * will change as the index grows and contracts.  The pointer to the root node
 * is stored in the index file at an offset equal to the index length
 * (i.e. at the beginning of the second index block).
 *---------------------------------------------------------------------------*/

#include "cbtree.h"

long getroot(fd, ptrloc)         /* get btree root record pointer */
int  fd;
long ptrloc; /* ptrloc is byte address where root ptr is stored */
{
   extern long lseek();
   long lbuff = 0L;    /* to avoid LC compiler warning */

   if (ptrloc < 1L)   /* root pointer is NEVER at base of file ! */
      ckerror(- CK_INDX, "128");
   if ((lbuff = lseek(fd, (long) ptrloc, 0)) == -1L)
      ckerror(- CK_SEEK, "47");
   if (read(fd, (char*)&lbuff, sizeof(lbuff) ) == ERR)
      ckerror(- CK_READ, "48");
   return(lbuff);
}


void getidxr(fd, recnum, idxblk, reclen, keylen)   /* get btree index record   */
int fd;
long recnum;
BTIDXBLK *idxblk;
unsigned reclen;
unsigned keylen;
{
   extern   long    lseek();
   extern   int   strlen();
   extern   void    free();
   extern   char *strcpy(), *strnncpy(), *calloc(), *strsave();
   char   *buff;
   uword  *uptr;
   long   *lptr;
   char   *cptr;
   register uword u;
#if NOALIGN
   uword len;
#endif

   if ((buff = calloc(reclen + 1, SZCHAR)) == NULL)   /* main buffer */
   {
#if 01
printf("\ngetbtree.c:giveidxr: can't get %u bytes of memory", reclen + 1);
#endif
      ckerror(- CK_NOMEM, "49");
   }
#if 0
printf("\ngetbtree: reading record # %lu", recnum);
#endif

   /* locate maint. info.    */
   if (btseek(fd, recnum, reclen) == -1L)
      ckerror(- CK_SEEK, "50");
   if (read(fd, buff, (int)reclen) == ERR)
      ckerror(- CK_READ, "51");

   /* use pointers to cast data types from buff into struct. */
   uptr = (uword *)buff;            /* start with the unsigned words */
   idxblk->blktype = *uptr++;
   idxblk->cellicnt = *uptr++;
#if 0
printf("\ngetbtree0: idxblk->cellicnt = %x", idxblk->cellicnt);
#endif

   lptr = (long *)uptr;           /* now do the longs */
   idxblk->fwdpage = *lptr++;
   idxblk->bwdpage = *lptr++;
#if NOALIGN
   cptr = (char *)(lptr + 1);              /* key is next to pointer */
#else
   cptr = (char *)(lptr + idxblk->cellicnt);  /* ptr's together, then keys */
#endif
   for (u = 0; u < idxblk->cellicnt; ++u)
   {
#if 0
printf("\ngetbtree1: cptr = %lx, buff = %lx, cptr - buff = %x, reclen = %x",
      cptr, buff, cptr - buff, reclen);
#endif
      idxblk->btpage[u].btptr = *lptr++;
      if (idxblk->blkalloc > u)          /* memory already allocated    */
         strnncpy(idxblk->btpage[u].skeynme, cptr, keylen);
      else
      {
         idxblk->btpage[u].skeynme = strsave(cptr, keylen); /* alloc mem */
         idxblk->blkalloc++;
      }
/* Low-byte High-byte order */
#if NOALIGN
      cptr += (len = strlen(cptr) + 1) + sizeof(long);   /* skip over long */
      lptr = (long *)((char *)lptr + len);               /* skip over str    */
#else   /* Low-byte High-byte order */
      cptr += strlen(cptr) + 1;    /* point to next string location */
#endif

#if 0
printf("\ngetbtree2: cptr = %lx, buff = %lx, cptr - buff = %x, reclen = %x",
      cptr, buff, cptr - buff, reclen);
printf("\ngetbtree: btpage[%u].skeynme = '%s', .btptr = %lu",
      u, idxblk->btpage[u].skeynme, idxblk->btpage[u].btptr);
#endif
      if ((cptr - buff) - sizeof(long) > reclen)
         ckerror(- CK_INDX, "81");
   }
   FREE(buff);
}

/* strsave(): alloc n bytes for saving string s and return pointer to s */

char *strsave(s, n)
char *s;
uint    n;
{
   extern char *strnncpy(), *calloc();
   char *p;

   if ((p = calloc(n + 1, SZCHAR)) == NULL)
      ckerror(- CK_NOMEM, "56");

   strnncpy(p, s, n);
   return (p);
}

